package clews.data;

import java.util.ArrayList;

import clews.env.Environment;
import clews.env.Utils;
import clews.gui.view.ConstraintView;

public class Constraint 
extends DataObject<ConstraintView>
{
	protected Specification specification;
	protected String name;
	protected String errorMessage;
	protected boolean active;
	protected ArrayList<ConstraintPart> lhs;
	protected ArrayList<ConstraintPart> rhs;
	protected Association reduced;
	protected int upperBound, plhs, prhs, prod;
	
	public Constraint(Specification s)
	{
		specification = s;
		name = "Constraint"+getId();
		active = true;
		lhs = new ArrayList<ConstraintPart>();
		rhs = new ArrayList<ConstraintPart>();
	}
	
	public String toString()
	{
		if (isActive())
			return getName();
		return "["+getName()+"]";
	}	

	public boolean computeResultingInterval()
	{
		if ((lhs.size() == 0 || rhs.size() == 0)||
		    (lhs.get(0).association.getTypeLeft() != rhs.get(0).association.getTypeLeft()))
		{
			errorMessage = "Left hand side and right hand side do not match";
			return false;
		}
		
		int[] b1, b2;
		b1 = b2 = null;

		if (lhs.get(0).association.getTypeLeft() == AssociationAttribute.NON_UNIQUE)			
		{
			b1 = computeBoundsNonUnique(lhs);
			plhs = prod;
			b2 = computeBoundsNonUnique(rhs);
			prhs = prod;
		}
		if (lhs.get(0).association.getTypeLeft() == AssociationAttribute.UNIQUE)
		{
			b1 = computeBoundsUnique(lhs);
			plhs = prod;
			b2 = computeBoundsUnique(rhs);
			prhs = prod;
		}
		
		int[] inter = new int[4];
		inter[0] = Math.max(b1[0], b2[0]);
		inter[1] = Math.min(b1[1], b2[1]);
		inter[2] = Math.max(b1[2], b2[2]);
		inter[3] = Math.min(b1[3], b2[3]);
		
		if (inter[0] > inter[1] || inter[2] > inter[3])
		{
			errorMessage = "The left hand side is of type "+"["+b1[0]+".."+b1[1]+"] --- ["+b1[2]+".."+b1[3]+"]"+" and\n"+
						  "the right hand side is of type "+"["+b2[0]+".."+b2[1]+"] --- ["+b2[2]+".."+b2[3]+"].\n"+
						  "Therefore the intersection is not satisfiable";
			
			return false;
		}
		System.out.println(inter[0]+" "+inter[1]+" "+inter[2]+" "+inter[3]);
		
		assignConstraint(lhs);
		assignConstraint(rhs);
		upperBound = inter[3];
		return true;		
	}
	
	public String getErrorMessage()
	{
		return errorMessage;
	}
	
	public int getLowerProduct(ConstraintPart cp)
	{
		if (lhs.contains(cp))
			return plhs;
		if (rhs.contains(cp))
			return prhs;
		return -1;
	}
	
	public int getUpperBound()
	{
		return upperBound;
	}
	
	public void removeFromAssociations()
	{
		for (ConstraintPart cp : lhs)
			cp.association.deleteConstraint(cp);
		for (ConstraintPart cp : rhs)
			cp.association.deleteConstraint(cp);
	}
	
	protected void assignConstraint(ArrayList<ConstraintPart> parts)
	{
		for (ConstraintPart cp : parts)
			cp.association.addConstraint(cp);
	}
	
	protected int maxMul(int a, int b)
	{
		if (a == Integer.MAX_VALUE)
			return a;
		if (b == Environment.CARDINALITY_STAR)
			return Integer.MAX_VALUE;
		return a*b;
	}
	
	public int[] computeBoundsNonUnique(ArrayList<ConstraintPart> parts) // also works for Unique
	{
		int[] bounds = new int[4];
		for (int i=0;i<4;i++)
			bounds[i] = 1;
		prod = 1;
		for (ConstraintPart cp : parts){
			if (!cp.inv)
			{
				bounds[0] *= cp.association.getLeft().getMin();
				bounds[1] = maxMul(bounds[1],cp.association.getLeft().getMax());
				bounds[2] *= cp.association.getRight().getMin();
				bounds[3] = maxMul(bounds[3],cp.association.getRight().getMax());
				prod *= cp.association.getRight().getMin();
			}
			else
			{
				bounds[2] *= cp.association.getLeft().getMin();
				bounds[3] = maxMul(bounds[3],cp.association.getLeft().getMax());
				bounds[0] *= cp.association.getRight().getMin();
				bounds[1] = maxMul(bounds[3],cp.association.getRight().getMax());
				prod *= cp.association.getLeft().getMin();
			}
		}
		return bounds;
	}
	
	public int[] computeBoundsUnique(ArrayList<ConstraintPart> parts) 
	{
		int[] bounds = new int[4];
		for (int i=0;i<4;i++)
			bounds[i] = 1;
		prod = 1;
		ConstraintPart cp;
		for (int i=0;i<parts.size();i++)
		{
			cp = parts.get(i);
			if (!cp.inv)
			{
				if (i == 0){
					bounds[0] *= cp.association.getLeft().getMin();
					bounds[2] *= cp.association.getRight().getMin();
				}else{
					bounds[0] *= Utils.sign(cp.association.getLeft().getMin());
					bounds[2] *= Utils.sign(cp.association.getRight().getMin());
				}
				bounds[1] = maxMul(bounds[1],cp.association.getLeft().getMax());				
				bounds[3] = maxMul(bounds[3],cp.association.getRight().getMax());
				prod *= cp.association.getRight().getMin();
			}
			else
			{
				if (i == 0){
					bounds[2] *= cp.association.getLeft().getMin();
					bounds[0] *= cp.association.getRight().getMin();
				}else{
					bounds[2] *= Utils.sign(cp.association.getLeft().getMin());
					bounds[0] *= Utils.sign(cp.association.getRight().getMin());					
				}				
				bounds[3] = maxMul(bounds[3],cp.association.getLeft().getMax());				
				bounds[1] = maxMul(bounds[1],cp.association.getRight().getMax());
				prod *= cp.association.getLeft().getMin();
			}
		}
		return bounds;
	}
	
	public boolean match(ConstraintPart cp1, ConstraintPart cp2)
	{
		if (cp1.inv == cp2.inv)
		{
			if (cp1.inv)
				return (cp1.association.getLeft().assclass.getId() == cp2.association.getRight().assclass.getId()); 
			else
				return (cp1.association.getRight().assclass.getId() == cp2.association.getLeft().assclass.getId());
		}
		else
		{
			if (cp1.inv)
				return (cp1.association.getLeft().assclass.getId() == cp2.association.getLeft().assclass.getId());
			else				
				return (cp1.association.getRight().assclass.getId() == cp2.association.getRight().assclass.getId());
		}			
	}		
	
	public ArrayList<ConstraintPart> getLhs()
	{
		return lhs;
	}
	
	public ArrayList<ConstraintPart> getRhs()
	{
		return rhs;
	}
	
	public void clear()
	{
		lhs.clear();
		rhs.clear();
	}
	
	public String getName()
	{
		return name;
	}
	
	public void setName(String s)
	{
		name = s;
	}
	
	public Specification getSpecification()
	{
		return specification;
	}
	
	public boolean isActive()
	{
		return active;
	}
	
	public void setActive(boolean a)
	{
		active = a;
	}
	
	public void addLhsPart(Association a, boolean inv)
	{
		lhs.add(new ConstraintPart(this,a, inv));
	}
	
	public void addRhsPart(Association a, boolean inv)
	{
		rhs.add(new ConstraintPart(this,a, inv));
	}
	
	public class ConstraintPart
	{
		public boolean inv;
		public Association association;
		public Constraint constraint;
		
		public ConstraintPart(Constraint c, Association a, boolean i)
		{
			constraint = c;
			association = a;
			inv = i;
		}
	}
}
