package clews.export;

import java.util.HashMap;

import net.sf.javailp.Constraint;
import net.sf.javailp.Linear;
import net.sf.javailp.OptType;
import net.sf.javailp.Problem;
import net.sf.javailp.Result;
import net.sf.javailp.Solver;
import net.sf.javailp.SolverFactory;
import net.sf.javailp.SolverFactoryLpSolve;
import net.sf.javailp.VarType;
import clews.MainFrame;
import clews.data.Association;
import clews.data.AssociationEnd;
import clews.data.AssociationAttribute;
import clews.data.Class;
import clews.data.Specification;
import clews.env.Environment;

public class SpecificationILPExport {

	protected Specification diagram;
	protected Problem problem;
	protected Result intermediate, finalres;
	protected HashMap<Constraint,String> added;
	protected HashMap<clews.data.Class, Integer> minObjects;
	
	public SpecificationILPExport(Specification d)
	{
		this(d,null);
	}
	
	public SpecificationILPExport(Specification d, HashMap<clews.data.Class, Integer> min)
	{
		diagram = d;
		intermediate = null;
		finalres = null;
		minObjects = min;
		problem = new Problem();
	}

	public Problem export()
	{
		problem = new Problem();
		for (Class c : diagram.getClasses())
			transformClass(c);
		for (Association a : diagram.getAssociations())
			transformAssociation(a);
		return problem;
	}
	
	protected void transformClass(Class c)
	{		
		if (minObjects != null || c.getMinObj() > 0){
			Linear l = new Linear();
			l.add(1, getName(c));
			if (minObjects != null)
				problem.add(l, ">=", Math.max(c.getMinObj(),minObjects.get(c).intValue()));
			else				
				problem.add(l, ">=", c.getMinObj());			
		}
		
		if (c.getMaxObj() != Environment.CARDINALITY_STAR)
		{
			Linear l = new Linear();
			l.add(1, getName(c));
			problem.add(l, "<=", c.getMaxObj());
		}
	}
	
	protected void addConditiondalInequalities()
	{
		added = new HashMap<Constraint, String>();
		OptType save = problem.getOptType();
		export();		
		setObjective(save);
		intermediate = null;
	}
	
	public Result solve()
	{
		SolverFactory factory = new SolverFactoryLpSolve(); // use lp_solve
		factory.setParameter(Solver.VERBOSE, 0);
		factory.setParameter(Solver.TIMEOUT, 100); // set timeout to 100 seconds
		
		finalres = null;
		try
		{
			int cc;
			do
			{
				cc = problem.getConstraintsCount(); 
				Solver solver = factory.get(); 
				finalres = intermediate = solver.solve(problem);
				addConditiondalInequalities();
			}while (cc != problem.getConstraintsCount());		
		}catch (Exception e)
		{
			MainFrame.getInstance().getTextOutput().setText(
					MainFrame.getInstance().getTextOutput().getText()+"\n"+
					"No ILP-Solver found!");
		}
		return finalres;
	}
	
	protected Linear transformAssociation(Association assoc)
	{
		Linear linear = new Linear();
		AssociationAttribute tl = assoc.getTypeLeft(),tr = assoc.getTypeRight();
		AssociationEnd l=assoc.getLeft(),r=assoc.getRight();
		if (tl == tr)
		{
			if (tl == AssociationAttribute.UNIQUE)
				addUniqueUnique(l,r);
			else if (tl == AssociationAttribute.NON_UNIQUE)
				addNonUniqueNonUnique(l,r);
			else if (tl == AssociationAttribute.ALL_SAME)
				addAllSameAllSame(l,r);
		}
		else
		{
			if (tl == AssociationAttribute.UNIQUE)
			{
				if (tr == AssociationAttribute.NON_UNIQUE)
					addNonUniqueUnique(r,l);
				else if (tr == AssociationAttribute.ALL_SAME)
					addAllSameUnique(r,l);
			} 
			else if (assoc.getTypeLeft() == AssociationAttribute.NON_UNIQUE)
			{
				if (tr == AssociationAttribute.UNIQUE)
					addNonUniqueUnique(l,r);
				else if (tr == AssociationAttribute.ALL_SAME)
					addAllSameNonUnique(r, l);					
			}
			else if (assoc.getTypeLeft() == AssociationAttribute.ALL_SAME)
			{
				if (tr == AssociationAttribute.UNIQUE)
					addAllSameUnique(l,r);
				else if (tr == AssociationAttribute.NON_UNIQUE)
					addAllSameNonUnique(l,r);
			}		
		}
		return linear;
	}
	
	public void printProblem()
	{
		String text = "";
		for (Constraint c : problem.getConstraints())
		{
		//	System.out.println(c);
			if (added.containsKey(c))
			{
				text = text + added.get(c);
			}
			text = text + c + "\n";
		}
		MainFrame.getInstance().getTextOutput().setText(text);
	}

	protected String getName(AssociationEnd ae)
	{
		return getName(ae.assclass);
	}
	
	protected String getName(Class c)
	{
		return makeVar(c.getName());
	}
	
	public int getResultVal(Class c)
	{
		return finalres.get(getName(c)).intValue();
	}
	
	public static String makeVar(String name)
	{
		return "|"+name+"|";
	}
	
	protected int getIntermediateObjects(AssociationEnd u)
	{
		if (intermediate != null)
			return intermediate.get(getName(u)).intValue();
		return 0;		
	}
	
	protected void addNonUnique(AssociationEnd u, AssociationEnd v)
	{
		if (v.getMax() != Environment.CARDINALITY_STAR)
		{
			Linear l = new Linear();			
			l.add(u.getMin(), getName(v)); 
			l.add(-v.getMax(), getName(u));
			problem.add(l, "<=", 0);
		}					
	}
	
	protected String getConditionalString(AssociationEnd u, int card)	
	{
		return getName(u)+" > "+card + "   -->   ";
	}
	
	protected void addUnique(AssociationEnd u, AssociationEnd v)
	{
		if (getIntermediateObjects(u) > 0)
		{
			Linear l = new Linear();
			l.add(1, getName(v));
			Constraint c = new Constraint(l, ">=", v.getMin());
			problem.add(c);			
			added.put(c, getConditionalString(u,0));
		}
	}
	
	protected void addNonUniqueNonUnique(AssociationEnd u, AssociationEnd v)
	{
		addNonUnique(u, v);
		addNonUnique(v, u);
	}
	
	protected void addUniqueUnique(AssociationEnd u, AssociationEnd v)
	{
		addNonUniqueNonUnique(u,v);
		addUnique(u,v);
		addUnique(v,u);
	}
	
	protected void addNonUniqueUnique(AssociationEnd u, AssociationEnd v)
	{		
		addNonUnique(v, u);
		addUnique(u,v);
		
		if (u.getMin() > 0 && v.getMax() != Environment.CARDINALITY_STAR)
		{
			Linear l = new Linear();
			l.add(1,getName(v));
			l.add(-v.getMax(), getName(u));
			problem.add(l, "<=", 0);	
		}
	}
	
	protected void addAllSameAllSame(AssociationEnd u, AssociationEnd v)
	{
		if (u.getMin() <= v.getMax() && v.getMin() <= u.getMax())
		{
			Linear l;
			if (u.getMin() > 0)
			{
				l = new Linear();
				l.add(1,getName(v));
				l.add(-1,getName(u));
				problem.add(l, "<=", 0);
			}
			if (v.getMin() > 0)
			{
				l = new Linear();
				l.add(1,getName(u));
				l.add(-1,getName(v));
				problem.add(l, "<=", 0);		
			}
		}
	}

	protected void addAllSameUnique(AssociationEnd u, AssociationEnd v)
	{
		Linear l = new Linear();
		l.add(v.getMin(),getName(u));
		l.add(-1,getName(v));
		problem.add(l, "<=", 0);
		
		if (u.getMin() > 0)
		{
			l = new Linear();
			l.add(1,getName(v));
			l.add(v.getMax(),getName(u));
			problem.add(l, "<=", 0);
		}
	}
	
	protected void addAllSameNonUnique(AssociationEnd u, AssociationEnd v)
	{
		Linear l;
		if (u.getMax() > 0)
		{
			l = new Linear();
			l.add((v.getMin()+u.getMax()-1)/u.getMax(), getName(u));
			l.add(-1,getName(v));
			problem.add(l, "<=", 0);
		}
		if (u.getMin() > 0)
		{
			l = new Linear();
			l.add(1,getName(v));
			l.add(-(v.getMax()/u.getMin()),getName(u));
			problem.add(l, "<=", 0);			
		}		
	}
	
	protected void setVarTypes() {
		for (Object o : problem.getVariables())
			problem.setVarType(o, VarType.INT);	
	}
	
	public void setObjective(OptType type)
	{
		setVarTypes();
		Linear linear = new Linear();
		for (clews.data.Class c : diagram.getClasses())
		{
			Linear children=new Linear(),children2 = new Linear();
			for (AssociationEnd cae : c.getEnds())
			{
				AssociationEnd ae = cae.getOther();
				if (ae.type == AssociationAttribute.UNIQUE && cae.type == AssociationAttribute.UNIQUE){
//					if (ae.getMin() == 1 &&  ae.getMax() == 1)
//						children.add(ae.getMax(),makeVar(cae.assclass.getName()));
					if (ae.getMin() == ae.getMax() && cae.getMin() != cae.getMax()){
						children.add(ae.getMax(),makeVar(cae.assclass.getName()));
						children2.add(ae.getMax(),makeVar(cae.assclass.getName()));
					}
//					if (cae.getMin()==0 && cae.getMax()==1 && ae.getMin() == ae.getMax())
//						children.add(ae.getMax(),makeVar(cae.assclass.getName()));
//					if (cae.getMin() == 1 && cae.getMax() == 1)
//						parents.add(1,makeVar(cae.assclass.getName()));
				}
			}
			
			if (children.size() > 1 /*c.getEnds().size() > 2*/)			
			{
				if (c.getMinPartner() != 0){
					children.add(-c.getMinPartner(),makeVar(c.getName()));
					problem.add(children,">=",0);
				}
				if (c.getMaxParnter() != Environment.CARDINALITY_STAR){
					children2.add(-c.getMaxParnter(),makeVar(c.getName()));
					problem.add(children2,"<=",0); // TODO
				}
			}
/*
			if (c.getName().equals("Target"))
			{
				Linear tmp = new Linear();
				tmp.add(1,makeVar(c.getName()));
				problem.add(tmp,"<=",1);
				tmp = new Linear();
				tmp.add(1,makeVar("Source"));
				problem.add(tmp,"<=",1);
			}
*/
			linear.add(c.getWeight(),makeVar(c.getName()));
		}
		problem.setObjective(linear, type);

/*		
		for (clews.data.Class c : diagram.getClasses()) {
			linear = new Linear();
			linear.add(1, makeVar(c.getName()));
			problem.add(linear, ">=", 1);
		}
*/		
	}
}
