package clews.gui.diagram;

import java.awt.Dimension;

import java.awt.Frame;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;
import java.util.ArrayList;

import javax.swing.JLabel;
import javax.swing.JWindow;
import javax.swing.UIManager;

import clews.env.Environment;
import clews.gui.view.PickableObject;
import clews.gui.view.SelectableObject;

public abstract class Diagram<P extends PickableObject, S extends SelectableObject> 
implements AnyDiagram
{	
	protected ArrayList<PickableObject> dragList;
	protected PickableObject drag, selection;
	protected DiagramPanel panel;
	protected SelectableObject overObject;
	protected boolean hasSelection,repaintRequest;
	protected Point2D mp;
	
    protected JWindow toolTip;
    protected JLabel toolTipLabel;
    protected boolean hovering;
    protected AffineTransform viewTransform;
    protected AffineTransform printShift;



	public Diagram(DiagramPanel dp)
	{
		viewTransform = new AffineTransform();
		printShift = new AffineTransform();
		panel = dp;
		hasSelection = false;
		
		drag = null;
		repaintRequest = false;
		mp = new Point();
		
		dragList = new ArrayList<PickableObject>();
		
		toolTip = null;
	}
	
	public abstract ArrayList<P> getPickableObjects();
	public abstract ArrayList<S> getSelectableObjects();
	
	public void draw(Graphics2D g)
	{
		if (drag != null)
			drag.draw(g);
	}
		
	public boolean hasSelection()
	{
		return hasSelection;
	}
	
	public boolean isOverObject()
	{
		return (overObject!=null);
	}


	public void clearDragList()
	{
		for (PickableObject po : dragList)
			po.unmark();
		dragList.clear();	
	}
	
    protected void initToolTip() {
        toolTipLabel = new JLabel(" ");
        toolTipLabel.setOpaque(true);
        toolTipLabel.setBackground(UIManager.getColor("ToolTip.background"));
        toolTip = new JWindow(new Frame());
        toolTip.getContentPane().add(toolTipLabel);
    }

	
	public void select(Rectangle r)
	{
		dragList.clear();
		for (PickableObject po : getPickableObjects())
		{			
			if (r.contains(po.getBounds())) 
			{
				dragList.add(po);
				po.mark();
			}
		}		
	}
	
	public P getPickableForPos(Point2D p)
	{
		for (P po : getPickableObjects())
			if (po.contains(p))
				return po;			
		return null;
	}	
	
	public void clearSelection()
	{
		hasSelection = false; 
		if (selection != null){
			selection.setSelected(false);
			selection = null;
		}
	}
		
	@Override
	public void mouseClicked(MouseEvent e) {
		if (e.getButton() == MouseEvent.BUTTON1 && e.getClickCount() > 1 && overObject != null)
		{
			overObject.doubleClicked(e.getPoint());
		}
		
		if (e.getButton() == MouseEvent.BUTTON3 && overObject != null)
		{
			overObject.rightClicked(e.getPoint());
		}
		
/*		
		this.diagram.applyChanges();
		*/
		
	}

	@Override
	public void mouseEntered(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mouseExited(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}
	
	public abstract void connect(PickableObject po1, PickableObject po2);	

	@Override
	public void mousePressed(MouseEvent e) {
		if (toolTip != null)
			toolTip.setVisible(false);
		if (e.getButton() == MouseEvent.BUTTON1)
		{
			transformMousePos(e);
			if (Environment.IS_CONTROL_DOWN)
			{
				PickableObject po = getPickableForPos(mp);
				if (po != null)
				{
					if (selection == null)
					{
						selection = po;
						selection.setSelected(true);
					}
					else
					{
						connect(selection, po);
						clearSelection();
					}
				}
				
			}
			else
			{
				selection = null;

				drag = getPickableForPos(mp);
				if (drag != null)
				{
					if (!dragList.contains(drag))
					{
						for (PickableObject po : dragList)
							po.unmark();
						dragList.clear();
					}
					hasSelection = true;
					drag.clicked(mp);
					for (PickableObject po : dragList)
						po.clicked(mp);
				}				
			}
		}		
	}
	
	public void checkToolTip(PickableObject po, Point p)
	{		
	}

	public void clearToolTip()
	{
		if(!hovering && toolTip != null && toolTip.isVisible())
			toolTip.setVisible(false);
	}

	@Override
	public void mouseReleased(MouseEvent e) {
		if (drag != null)
		{
			hasSelection = false;
			drag.release();
			for (PickableObject po : dragList)
				po.release();
			dragList.clear();
		}
		if (!Environment.IS_CONTROL_DOWN)
			selection = null;
		drag = null;		
	}

	@Override
	public void mouseDragged(MouseEvent e) {
		if (drag != null)
		{
			transformMousePos(e);
			drag.updateLocation(mp);
			for (PickableObject po : dragList)
				po.updateLocation(mp);
			repaintRequest = true;
		}	

		if (!Environment.IS_CONTROL_DOWN)
			selection = null;
		repaint();				
	}

	@Override
	public void mouseMoved(MouseEvent e) {
		transformMousePos(e);
		overObject = null;
				
		hovering = false;
		for (PickableObject po : getPickableObjects()){
			if (po.contains(mp))
			{
				repaintRequest = repaintRequest | po.setOver(true);
				if (overObject == null)
					overObject = po;
				checkToolTip(po,e.getPoint());
			}
			else
			{
				if (!dragList.contains(po))
					repaintRequest = repaintRequest | po.setOver(false);
			}
		}		
		clearToolTip();
		
		for (SelectableObject so : getSelectableObjects())
		{	
			if (overObject == null && so.contains(mp))
			{
				repaintRequest = repaintRequest | so.setOver(true);
				overObject = so;
			}
			else
				repaintRequest = repaintRequest | so.setOver(false);
		}
		repaint();
	}
	
	public AffineTransform getPrintShift()
	{
		return printShift;
	}
	
	public Dimension setUpPrint()
	{
		int minx,maxx,miny,maxy;
		
		minx = miny = Integer.MAX_VALUE;
		maxx = maxy = Integer.MIN_VALUE;
		for (PickableObject cv : getPickableObjects())
		{			 
			minx = Math.min(minx,cv.getPrintBounds().x);
			miny = Math.min(miny,cv.getPrintBounds().y);
			maxx = Math.max(maxx, cv.getPrintBounds().x + cv.getPrintBounds().width);
			maxy = Math.max(maxy, cv.getPrintBounds().y + cv.getPrintBounds().height);
		}
		for (SelectableObject cv : getSelectableObjects())
		{			 
			minx = Math.min(minx,cv.getPrintBounds().x);
			miny = Math.min(miny,cv.getPrintBounds().y);
			maxx = Math.max(maxx, cv.getPrintBounds().x + cv.getPrintBounds().width);
			maxy = Math.max(maxy, cv.getPrintBounds().y + cv.getPrintBounds().height);
		}
		printShift = AffineTransform.getTranslateInstance(-minx+Environment.PRINT_SPACING, -miny+Environment.PRINT_SPACING);
			
		return new Dimension(maxx-minx+2*+Environment.PRINT_SPACING,
				maxy-miny+2*+Environment.PRINT_SPACING);
	}
	
	public Point2D transformMousePos(MouseEvent e)
	{
		getInverseTransform().transform(e.getPoint(), mp);
		return mp;
	}
	
	public void setRepaint()
	{
		repaintRequest = true;
	}
	
	public void repaint()
	{
		if (repaintRequest)
		{
			panel.repaint();
			repaintRequest = false;
		}
	}
	
	public AffineTransform getTransform()
	{
		return viewTransform;
	}	
	
	public void setTransform(AffineTransform at)
	{
		viewTransform = at;
	}
	
	public void setTransformParameter(double scale, double transx, double transy)
	{
		viewTransform = AffineTransform.getTranslateInstance(transx, transy);
		viewTransform.scale(scale, scale);
	}

	public AffineTransform getInverseTransform()
	{
		try {
			return getTransform().createInverse();
		} catch (NoninvertibleTransformException e1) {
			e1.printStackTrace();
			return null;	
		}
	}

}
